"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.StoragePlugin = void 0;
const plugin_1 = require("appium/plugin");
const storage_1 = require("./storage");
const lodash_1 = __importDefault(require("lodash"));
const support_1 = require("@appium/support");
const lru_cache_1 = require("lru-cache");
const ws_1 = __importDefault(require("ws"));
const node_stream_1 = require("node:stream");
const utils_1 = require("./utils");
const log = support_1.logger.getLogger('StoragePlugin');
let SHARED_STORAGE = null;
const STORAGE_PREFIX = '/storage';
const WS_TTL_MS = 5 * 60 * 1000;
const STORAGE_HANDLERS = {};
const STORAGE_ADDITIONS_CACHE = new lru_cache_1.LRUCache({
    max: 20,
    ttl: WS_TTL_MS,
    dispose: (f) => f(),
});
class StoragePlugin extends plugin_1.BasePlugin {
    static async updateServer(expressApp, httpServer) {
        const buildHandler = (methodName) => async (req, res) => {
            let status = 200;
            let body;
            try {
                const value = await STORAGE_HANDLERS[methodName](req, httpServer);
                body = { value: value ?? null };
            }
            catch (e) {
                [status, body] = (0, utils_1.toW3cResponseError)(e);
            }
            log.debug(`Responding to ${methodName} with ${lodash_1.default.truncate(JSON.stringify(body.value), { length: 200 })}`);
            res.set('content-type', 'application/json; charset=utf-8');
            res.status(status).send(body);
        };
        expressApp.post(`${STORAGE_PREFIX}/add`, buildHandler(STORAGE_HANDLERS.addStorageItem.name));
        expressApp.get(`${STORAGE_PREFIX}/list`, buildHandler(STORAGE_HANDLERS.listStorageItems.name));
        expressApp.post(`${STORAGE_PREFIX}/reset`, buildHandler(STORAGE_HANDLERS.resetStorage.name));
        expressApp.post(`${STORAGE_PREFIX}/delete`, buildHandler(STORAGE_HANDLERS.deleteStorageItem.name));
    }
}
exports.StoragePlugin = StoragePlugin;
STORAGE_HANDLERS.addStorageItem = async function addStorageItem(req, httpServer) {
    const itemOptions = (0, storage_1.requireValidItemOptions)(parseRequestArgs(req, ['name', 'sha1']));
    const [stream, events] = prepareWebSockets(httpServer, itemOptions);
    return {
        ws: {
            stream,
            events,
        },
        ttlMs: WS_TTL_MS,
    };
};
STORAGE_HANDLERS.listStorageItems = async function listStorageItems() {
    return await executeStorageMethod(async (storage) => await storage.list());
};
STORAGE_HANDLERS.deleteStorageItem = async function deleteStorageItem(req) {
    return await executeStorageMethod(async (storage) => await storage.delete(parseRequestArgs(req, ['name']).name));
};
STORAGE_HANDLERS.resetStorage = async function resetStorage() {
    await executeStorageMethod(async (storage) => await storage.reset());
};
async function executeStorageMethod(method) {
    const storage = await getStorageSingleton();
    return await method(storage);
}
function parseRequestArgs(req, requiredKeys) {
    if (!lodash_1.default.isPlainObject(req.body)) {
        throw new storage_1.StorageArgumentError(`The request body must be a valid JSON object`);
    }
    for (const key of requiredKeys) {
        if (!lodash_1.default.keys(req.body).includes(key)) {
            throw new storage_1.StorageArgumentError(`The required argument '${key}' is missing (expected ${JSON.stringify(requiredKeys)})`);
        }
    }
    return req.body;
}
function prepareWebSockets(httpServer, itemOptions) {
    const commonPathname = `${STORAGE_PREFIX}/add/${itemOptions.sha1}`;
    const streamPathname = `${commonPathname}/stream`;
    const eventsPathname = `${commonPathname}/events`;
    if (!lodash_1.default.isEmpty(httpServer.getWebSocketHandlers(streamPathname))) {
        return [streamPathname, eventsPathname];
    }
    const streamServer = new ws_1.default.Server({
        noServer: true,
    });
    const eventsServer = new ws_1.default.Server({
        noServer: true,
    });
    const signaler = new node_stream_1.EventEmitter();
    const streamDoneCallback = () => {
        log.debug(`Unmounting stream and events web sockets at ${commonPathname}`);
        httpServer.removeWebSocketHandler(streamPathname);
        httpServer.removeWebSocketHandler(eventsPathname);
        setTimeout(() => {
            streamServer.close();
            eventsServer.close();
            signaler.removeAllListeners();
        }, 100);
    };
    STORAGE_ADDITIONS_CACHE.set(itemOptions.sha1, streamDoneCallback);
    eventsServer.on('connection', async (wsUpstream) => {
        signaler.on('status', (value) => wsUpstream.send(JSON.stringify(value)));
    });
    eventsServer.on('error', (e) => {
        log.info(`The ${eventsPathname} web socket server has notified about an error: ${e.message}`);
    });
    streamServer.on('connection', async (wsUpstream) => {
        log.info(`Starting a new server storage upload of '${itemOptions.name}' at ${streamPathname}`);
        const storage = await getStorageSingleton();
        try {
            await storage.add(itemOptions, wsUpstream);
            const successEvent = {
                value: {
                    success: true,
                    ...itemOptions,
                }
            };
            log.debug(`Notifying about the successful addition of '${itemOptions.name}' to the server storage`);
            signaler.emit('status', successEvent);
            STORAGE_ADDITIONS_CACHE.delete(itemOptions.sha1);
        }
        catch (e) {
            log.debug(`Notifying about a failure while adding '${itemOptions.name}' to the server storage`);
            // in case of a failure we do not want to close the server yet
            // in anticipation of a retry
            log.error(e);
            const [, errorBody] = (0, utils_1.toW3cResponseError)(e);
            signaler.emit('status', errorBody);
        }
    });
    streamServer.on('error', (e) => {
        log.info(`The ${streamPathname} web socket server has notified about an error: ${e.message}`);
    });
    httpServer.addWebSocketHandler(streamPathname, streamServer);
    httpServer.addWebSocketHandler(eventsPathname, eventsServer);
    return [streamPathname, eventsPathname];
}
const getStorageSingleton = lodash_1.default.memoize(async () => {
    let storageRoot;
    let shouldPreserveRoot = false;
    let shouldPreserveFiles = false;
    if (process.env.APPIUM_STORAGE_ROOT) {
        storageRoot = process.env.APPIUM_STORAGE_ROOT;
        shouldPreserveRoot = shouldPreserveFiles = await support_1.fs.exists(storageRoot);
        log.info(`Set '${storageRoot}' as the server storage root folder`);
    }
    else {
        storageRoot = await support_1.tempDir.openDir();
        log.info(`Created '${storageRoot}' as the temporary server storage root folder`);
    }
    if (process.env.APPIUM_STORAGE_KEEP_ALL) {
        shouldPreserveFiles = ['true', '1', 'yes'].includes(lodash_1.default.toLower(process.env.APPIUM_STORAGE_KEEP_ALL));
    }
    if (shouldPreserveFiles) {
        log.info(`All server storage items will be always preserved unless deleted explicitly`);
    }
    else {
        log.info(`All server storage items will be cleaned up automatically from '${storageRoot}' after ` +
            `Appium server termination`);
    }
    SHARED_STORAGE = new storage_1.Storage(storageRoot, shouldPreserveRoot, shouldPreserveFiles, log);
    await SHARED_STORAGE.reset();
    return SHARED_STORAGE;
});
process.once('exit', () => {
    SHARED_STORAGE?.cleanupSync();
});
exports.default = StoragePlugin;
//# sourceMappingURL=plugin.js.map